import _accupy
import numpy

from .sums import fsum, ksum

# def dot2(x, y, prod2=prod2_fma):
#     '''Algorithm 5.3. Dot product in twice the working precision.
#     in <https://doi.org/10.1137/030601818>.
#     '''
#     p, s = prod2(x[0], y[0])
#     n = len(x)
#     for k in range(1, n):
#         h, r = prod2(x[k], y[k])
#         p, q = knuth_sum(p, h)
#         s += q+r
#     return p + s


def kdot(x, y, K=2):
    """Algorithm 5.10. Dot product algorithm in K-fold working precision,
    K >= 3.
    """
    xx = x.reshape(-1, x.shape[-1])
    yy = y.reshape(y.shape[0], -1)

    xx = numpy.ascontiguousarray(xx)
    yy = numpy.ascontiguousarray(yy)

    r = _accupy.kdot_helper(xx, yy).reshape((-1,) + x.shape[:-1] + y.shape[1:])
    return ksum(r, K - 1)


def fdot(x, y):
    """Algorithm 5.10. Dot product algorithm in K-fold working precision,
    K >= 3.
    """
    xx = x.reshape(-1, x.shape[-1])
    yy = y.reshape(y.shape[0], -1)

    xx = numpy.ascontiguousarray(xx)
    yy = numpy.ascontiguousarray(yy)

    r = _accupy.kdot_helper(xx, yy).reshape((-1,) + x.shape[:-1] + y.shape[1:])
    return fsum(r)



import dufte
import matplotlib.pyplot as plt
import numpy
import perfplot
import pytest

import accupy


@pytest.mark.parametrize("cond", [1.0, 1.0e15])
def test_kdot2(cond):
    x, y, ref, _ = accupy.generate_ill_conditioned_dot_product(100, cond)
    assert abs(accupy.kdot(x, y, K=2) - ref) < 1.0e-13 * abs(ref)


@pytest.mark.parametrize("cond", [1.0, 1.0e15, 1.0e30])
def test_kdot3(cond):
    x, y, ref, _ = accupy.generate_ill_conditioned_dot_product(100, cond)
    assert abs(accupy.kdot(x, y, K=3) - ref) < 1.0e-13 * abs(ref)


@pytest.mark.parametrize("cond", [1.0, 1.0e15, 1.0e30, 1.0e38])
def test_fdot(cond):
    x, y, ref, _ = accupy.generate_ill_conditioned_dot_product(100, cond)
    assert abs(accupy.fdot(x, y) - ref) < 1.0e-13 * abs(ref)


def test_accuracy_comparison_illcond(filename=None, target_cond=None):
    plt.style.use(dufte.style)

    if target_cond is None:
        target_cond = [10 ** k for k in range(2)]

    kernels = [
        numpy.dot,
        lambda x, y: accupy.kdot(x, y, K=2),
        lambda x, y: accupy.kdot(x, y, K=3),
        accupy.fdot,
    ]
    labels = ["numpy.dot", "accupy.kdot[2]", "accupy.kdot[3]", "accupy.fdot"]
    data = numpy.empty((len(target_cond), len(kernels)))
    condition_numbers = numpy.empty(len(target_cond))
    numpy.random.seed(0)
    for k, tc in enumerate(target_cond):
        x, y, ref, C = accupy.generate_ill_conditioned_dot_product(1000, tc)
        condition_numbers[k] = C
        data[k] = [abs(kernel(x, y) - ref) / abs(ref) for kernel in kernels]

    # sort
    s = numpy.argsort(condition_numbers)
    condition_numbers = condition_numbers[s]
    data = data[s]

    for label, d in zip(labels, data.T):
        plt.loglog(condition_numbers, d, label=label)

    dufte.legend()
    plt.xlabel("condition number")
    plt.ylabel("relative error")

    # plt.show()
    # <https://stackoverflow.com/a/10154763/353337>
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_speed_comparison1(filename=None, n_range=None):
    plt.style.use(dufte.style)

    if n_range is None:
        n_range = [2 ** k for k in range(2)]

    numpy.random.seed(0)
    perfplot.plot(
        setup=lambda n: (numpy.random.rand(n, 100), numpy.random.rand(100, n)),
        kernels=[
            lambda xy: numpy.dot(*xy),
            lambda xy: accupy.kdot(*xy, K=2),
            lambda xy: accupy.kdot(*xy, K=3),
            lambda xy: accupy.fdot(*xy),
        ],
        labels=["numpy.dot", "accupy.kdot[2]", "accupy.kdot[3]", "accupy.fdot"],
        n_range=n_range,
        xlabel="n",
    )
    plt.title("dot(random(n, 100), random(100, n))")

    # plt.show()
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_speed_comparison2(filename=None, n_range=None):
    if n_range is None:
        n_range = [2 ** k for k in range(2)]

    numpy.random.seed(0)
    perfplot.plot(
        setup=lambda n: (numpy.random.rand(100, n), numpy.random.rand(n, 100)),
        kernels=[
            lambda xy: numpy.dot(*xy),
            lambda xy: accupy.kdot(*xy, K=2),
            lambda xy: accupy.kdot(*xy, K=3),
            lambda xy: accupy.fdot(*xy),
        ],
        labels=["numpy.dot", "accupy.kdot[2]", "accupy.kdot[3]", "accupy.fdot"],
        n_range=n_range,
        xlabel="n",
        logx=True,
        logy=True,
    )
    plt.title("dot(random(100, n), random(n, 100))")
    # plt.show()

    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_discontiguous():
    x = numpy.random.rand(3, 10)
    y = numpy.random.rand(3, 10)
    accupy.kdot(x.T, y)
    accupy.fdot(x.T, y)


if __name__ == "__main__":
    # test_accuracy_comparison_illcond(
    #     "accuracy-dot.svg", [10 ** k for k in range(0, 37, 1)]
    # )
    # test_speed_comparison1(
    #     "speed-comparison-dot1.svg", n_range=[2 ** k for k in range(8)]
    # )
    test_speed_comparison2(
        "speed-comparison-dot2.svg", n_range=[2 ** k for k in range(8)]
    )



import dufte
import matplotlib.pyplot as plt
import numpy
import perfplot
import pytest

import accupy


@pytest.mark.parametrize("cond", [1.0, 1.0e15])
def test_ksum2(cond):
    p, ref, _ = accupy.generate_ill_conditioned_sum(100, cond)
    assert abs(accupy.ksum(p, K=2) - ref) < 1.0e-15 * abs(ref)


@pytest.mark.parametrize("cond", [1.0, 1.0e15, 1.0e30])
def test_ksum3(cond):
    p, ref, _ = accupy.generate_ill_conditioned_sum(100, cond)
    assert abs(accupy.ksum(p, K=3) - ref) < 1.0e-15 * abs(ref)


@pytest.mark.parametrize("cond", [1.0, 1.0e15, 1.0e30, 1.0e35])
def test_fsum(cond):
    p, ref, _ = accupy.generate_ill_conditioned_sum(100, cond)
    assert abs(accupy.fsum(p) - ref) < 1.0e-15 * abs(ref)


def test_accuracy_comparison_illcond(filename=None, target_conds=None):
    plt.style.use(dufte.style)

    if target_conds is None:
        target_conds = [10 ** k for k in range(1, 2)]

    kernels = [
        sum,
        numpy.sum,
        accupy.kahan_sum,
        lambda p: accupy.ksum(p, K=2),
        lambda p: accupy.ksum(p, K=3),
        accupy.fsum,
    ]
    labels = [
        "sum",
        "numpy.sum",
        "accupy.kahan_sum",
        "accupy.ksum[2]",
        "accupy.ksum[3]",
        "accupy.fsum",
    ]
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][: len(labels)]

    data = numpy.empty((len(target_conds), len(kernels)))
    condition_numbers = numpy.empty(len(target_conds))
    numpy.random.seed(0)
    for k, target_cond in enumerate(target_conds):
        p, ref, C = accupy.generate_ill_conditioned_sum(1000, target_cond)
        condition_numbers[k] = C
        data[k] = [abs(kernel(p) - ref) / abs(ref) for kernel in kernels]

    # sort
    s = numpy.argsort(condition_numbers)
    condition_numbers = condition_numbers[s]
    data = data[s]

    for label, color, d in zip(labels, colors, data.T):
        plt.loglog(condition_numbers, d, label=label, color=color)

    dufte.legend()
    plt.xlabel("condition number")
    plt.ylabel("relative error")
    # plt.gca().set_aspect(1.3)

    # plt.show()
    # <https://stackoverflow.com/a/10154763/353337>
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_speed_comparison1(filename=None, n_range=None):
    plt.style.use(dufte.style)

    if n_range is None:
        n_range = [2 ** k for k in range(2)]

    numpy.random.seed(0)
    perfplot.plot(
        setup=lambda n: numpy.random.rand(n, 100),
        kernels=[
            sum,
            lambda p: numpy.sum(p, axis=0),
            accupy.kahan_sum,
            lambda p: accupy.ksum(p, K=2),
            lambda p: accupy.ksum(p, K=3),
            accupy.fsum,
        ],
        labels=[
            "sum",
            "numpy.sum",
            "accupy.kahan_sum",
            "accupy.ksum[2]",
            "accupy.ksum[3]",
            "accupy.fsum",
        ],
        n_range=n_range,
        xlabel="n",
    )
    plt.title("Sum(random(n, 100))")

    # plt.show()
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_speed_comparison2(filename=None, n_range=None):
    plt.style.use(dufte.style)

    if n_range is None:
        n_range = [2 ** k for k in range(2)]

    numpy.random.seed(0)
    perfplot.plot(
        setup=lambda n: numpy.random.rand(100, n),
        kernels=[
            sum,
            lambda p: numpy.sum(p, axis=0),
            accupy.kahan_sum,
            lambda p: accupy.ksum(p, K=2),
            lambda p: accupy.ksum(p, K=3),
            accupy.fsum,
        ],
        labels=[
            "sum",
            "numpy.sum",
            "accupy.kahan_sum",
            "accupy.ksum[2]",
            "accupy.ksum[3]",
            "accupy.fsum",
        ],
        n_range=n_range,
        xlabel="n",
    )
    plt.title("Sum(random(100, n))")

    # plt.show()
    if filename:
        plt.savefig(filename, transparent=True, bbox_inches="tight")


def test_knuth_sum():
    a16 = numpy.float16(1.0e1)
    b16 = numpy.float16(1.0e-1)
    x16, y16 = accupy.knuth_sum(a16, b16)
    xy = numpy.float64(x16) + numpy.float64(y16)
    ab = numpy.float64(a16) + numpy.float64(b16)
    assert abs(xy - ab) < 1.0e-15 * ab


def test_decker_sum():
    a16 = numpy.float16(1.0e1)
    b16 = numpy.float16(1.0e-1)
    x16, y16 = accupy.decker_sum(a16, b16)
    xy = numpy.float64(x16) + numpy.float64(y16)
    ab = numpy.float64(a16) + numpy.float64(b16)
    assert abs(xy - ab) < 1.0e-15 * ab


def test_discontiguous():
    x = numpy.random.rand(3, 10).T
    accupy.ksum(x.T)
    accupy.fsum(x.T)


if __name__ == "__main__":
    # test_accuracy_comparison_illcond(
    #     "accuracy-sum.svg", [10 ** k for k in range(0, 37, 1)]
    # )
    # test_speed_comparison1("speed-comparison1.svg", n_range=[2 ** k for k in range(15)])
    test_speed_comparison2("speed-comparison2.svg", n_range=[2 ** k for k in range(15)])



from .__about__ import __version__
from .dot import fdot, kdot
from .ill_cond import generate_ill_conditioned_dot_product, generate_ill_conditioned_sum
from .sums import decker_sum, distill, fsum, kahan_sum, knuth_sum, ksum

__all__ = [
    "__version__",
    "kdot",
    "fdot",
    "generate_ill_conditioned_sum",
    "generate_ill_conditioned_dot_product",
    "knuth_sum",
    "decker_sum",
    "distill",
    "ksum",
    "fsum",
    "kahan_sum",
]



import math

import numpy
import pyfma
from mpmath import mp


def generate_ill_conditioned_sum(n, c, dps=100):
    # From <https://doi.org/10.1137/030601818>:
    # Ill-conditioned sums of length 2n are generated from dot products of
    # length n using Algorithm 3.3 (TwoProduct) and randomly permuting the
    # summands.
    x, y, _, C = generate_ill_conditioned_dot_product(n, c, dps)

    prod = x * y
    err = pyfma.fma(x, y, -prod)
    res = numpy.array([prod, err])

    out = numpy.random.permutation(res.flatten())

    def sum_exact(p):
        mp.dps = dps
        return mp.fsum(p)

    exact = sum_exact(out)
    # cond = sum_exact(numpy.abs(out)) / abs(exact)
    cond = C / 2

    return out, exact, cond


def generate_ill_conditioned_dot_product(n, c, dps=100):
    """n ... length of vector
    c ... target condition number
    """
    # Algorithm 6.1 from
    #
    # ACCURATE SUM AND DOT PRODUCT,
    # TAKESHI OGITA, SIEGFRIED M. RUMP, AND SHIN'ICHI OISHI.
    assert n >= 6
    n2 = round(n / 2)
    x = numpy.zeros(n)
    y = numpy.zeros(n)

    b = math.log2(c)
    # vector of exponents between 0 and b/2:
    e = numpy.rint(numpy.random.rand(n2) * b / 2).astype(int)
    # make sure exponents b/2 and 0 actually occur in e
    # vectors x,y
    e[0] = round(b / 2) + 1
    e[-1] = 0

    # generate first half of vectors x, y
    rx, ry = numpy.random.rand(2, n2)
    x[:n2] = (2 * rx - 1) * 2 ** e
    y[:n2] = (2 * ry - 1) * 2 ** e

    def dot_exact(x, y):
        mp.dps = dps
        # convert to list first, see
        # <https://github.com/fredrik-johansson/mpmath/pull/385>
        return mp.fdot(x.tolist(), y.tolist())

    # for i=n2+1:n and v=1:i,
    #     generate x_i, y_i such that (*) x(v)’*y(v) ~ 2^e(i-n2)
    # generate exponents for second half
    e = numpy.rint(numpy.linspace(b / 2, 0, n - n2)).astype(int)
    rx, ry = numpy.random.rand(2, n2)
    for i in range(n2, n):
        # x_i random with generated exponent
        x[i] = (2 * rx[i - n2] - 1) * 2 ** e[i - n2]
        # y_i according to (*)
        y[i] = (
            (2 * ry[i - n2] - 1) * 2 ** e[i - n2] - dot_exact(x[: i + 1], y[: i + 1])
        ) / x[i]

    x, y = numpy.random.permutation((x, y))
    # the true dot product rounded to nearest floating point
    d = dot_exact(x, y)
    # the actual condition number
    C = 2 * dot_exact(abs(x), abs(y)) / abs(d)

    return x, y, d, C



import math

import _accupy
import numpy


def knuth_sum(a, b):
    """Error-free transformation of the sum of two floating point numbers
    according to

    D.E. Knuth.
    The Art of Computer Programming: Seminumerical Algorithms, volume 2.
    Addison Wesley, Reading, Massachusetts, second edition, 1981.

    The underlying problem is that the exact sum a+b of two floating point
    number a and b is not necessarily a floating point number; for example if
    you add a very large and a very small number. It is however known that the
    difference between the best floating point approximation of a+b and the
    exact a+b is again a floating point number. This routine returns the sum
    and the error.

    Algorithm 3.1 in <https://doi.org/10.1137/030601818>.
    """
    x = a + b
    z = x - a
    y = (a - (x - z)) + (b - z)
    return x, y


def decker_sum(a, b):
    """Computationally equivalent to knuth_sum, but formally a bit cheaper.
    Only works for floats though (and not arrays), and the branch make it in
    fact less favorable in terms of actual speed.
    """
    x = a + b
    y = b - (x - a) if abs(a) > abs(b) else a - (x - b)
    return x, y


def distill(p, K):
    """Algorithm 4.3. Error-free vector transformation for summation.

    The vector p is transformed without changing the sum, and p_n is replaced
    by float(sum(p)). Kahan [21] calls this a 'distillation algorithm.'
    """
    q = p.reshape(p.shape[0], -1)
    for _ in range(K):
        _accupy.distill(q)
    return q.reshape(p.shape)


def ksum(p, K=2):
    """From

    T. Ogita, S.M. Rump, and S. Oishi.
    Accurate Sum and Dot Product,
    SIAM J. Sci. Comput., 26(6), 1955–1988 (34 pages).
    <https://doi.org/10.1137/030601818>.

    Algorithm 4.8. Summation as in K-fold precision by (K−1)-fold error-free
    vector transformation.
    """
    # Don't override the input data.
    q = p.copy()
    distill(q, K - 1)
    return numpy.sum(q[:-1], axis=0) + q[-1]


_math_fsum_vec = numpy.vectorize(math.fsum, signature="(m)->()")


def fsum(p):
    return _math_fsum_vec(p.T).T


def kahan_sum(p):
    """Kahan summation
    <https://en.wikipedia.org/wiki/Kahan_summation_algorithm>.
    """
    q = p.reshape(p.shape[0], -1)
    s = _accupy.kahan(q)
    return s.reshape(p.shape[1:])



# https://github.com/pybind/python_example
from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup

ext_modules = [
    Pybind11Extension(
        "_accupy",
        ["src/pybind11.cpp"],
        include_dirs=["/usr/include/eigen3/"],
    )
]

if __name__ == "__main__":
    setup(
        cmdclass={"build_ext": build_ext},
        ext_modules=ext_modules,
        zip_safe=False,
    )
